unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, ComCtrls;

const Sbase=10;
      Ssin60=9;
      Scos60=5;
      BaseX=-2;
      BaseY=0;
      MaxX=14;
      MaxY=17;
      Scircle=5;
      Tolerance=3;
      WinText='Super!';
      WinTextSize=48;

type
  TMainForm = class(TForm)
    PaintBox1: TPaintBox;
    BRestart: TButton;
    LTraps: TLabel;
    ETraps: TEdit;
    CTraps: TUpDown;
    Timer1: TTimer;
    procedure PaintBox1Paint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure BRestartClick(Sender: TObject);
    procedure CTrapsClick(Sender: TObject; Button: TUDBtnType);
    procedure Timer1Timer(Sender: TObject);
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
    Field:array[0..MaxY,0..MaxX]of boolean;
    Traps:integer;
    Dest,Pos:TPoint;
    procedure MakeField;
    procedure FindDest;
    procedure Draw(P:TPoint);
    function GetPoint(P:TPoint):TPoint;
    function InvGetPoint(P:TPoint):TPoint;
    procedure DrawWonText;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

procedure TMainForm.PaintBox1Paint(Sender: TObject);
var X,Y:integer;
begin
  for Y:=0 to MaxY do
    for X:=0 to MaxX do
      Draw(Point(X,Y));
end;

function TMainForm.GetPoint(P:TPoint):TPoint;
begin
  Result.X:=BaseX+(2*P.X+P.Y AND 1)*Ssin60;
  Result.Y:=BaseY+P.Y*(Scos60+Sbase);
end;

procedure TMainForm.Draw(P:TPoint);
var BX,BY:integer; D:TPoint;
begin
  with PaintBox1.Canvas do begin
    D:=GetPoint(P);
    BX:=D.X; BY:=D.Y;
    if (P.X=Dest.X) AND (P.Y=Dest.Y)
      then Brush.Color:=RGB(000+random(016),064+random(128),192+random(064))
    else if Field[P.Y,P.X]
      then Brush.Color:=RGB(120+random(016),128+random(016),128+random(016))
      else Brush.Color:=RGB(240+random(016),096+random(144),000+random(016));
    Polygon([Point(BX,BY-Sbase),Point(BX+Ssin60,BY-Scos60),Point(BX+Ssin60,BY+Scos60),
             Point(BX,BY+Sbase),Point(BX-Ssin60,BY+Scos60),Point(BX-Ssin60,BY-Scos60)]);
    if (P.X=Pos.X) AND (P.Y=Pos.Y) then begin
      Pen.Style:=psClear;
      Brush.Color:=RGB(000+random(064),192+random(064),000+random(016));
      Ellipse(BX-Scircle,BY-Scircle,BX+1+Scircle,BY+1+Scircle);
      Pen.Style:=psSolid; Pen.Color:=0;
    end;
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  randomize;
  Traps:=CTraps.Position;
  MakeField;
end;

procedure TMainForm.MakeField;
var X,Y,I:integer;
  procedure FindPosition;
  begin
    repeat
      X:=1+random(MaxX-1);
      Y:=1+random(MaxY-1);
    until Field[Y,X];
  end;
begin
  FillChar(Field,sizeof(Field),0);
  for Y:=1 to (MaxY-1) do
    for X:=1 to (MaxX-1) do
      Field[Y,X]:=true;
  for I:=1 to Traps do begin
    FindPosition; Field[Y,X]:=false;
  end;
  FindPosition; Pos.X:=X; Pos.Y:=Y;
  FindDest;
end;

procedure TMainForm.FindDest;
var X,Y:integer;
begin
  repeat
    X:=1+random(MaxX-1);
    Y:=1+random(MaxY-1);
  until Field[Y,X] AND not((X=Pos.X) AND (Y=Pos.Y));
  Dest.X:=X; Dest.Y:=Y;
end;

procedure TMainForm.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var   fPlayer,  fClick,  fRemove:TPoint;
    scrPlayer,scrClick,scrRemove:TPoint;
    Dist:integer;
begin
  fClick:=InvGetPoint(Point(X,Y));
  if Field[fClick.Y,fClick.X] then begin
    scrPlayer:=GetPoint(Pos);
    scrClick:=GetPoint(fClick);
    Dist:=round(sqrt(integer(sqr(scrPlayer.X-scrClick.X)+sqr(scrPlayer.Y-scrClick.Y))))
         -2*Ssin60;
    if abs(Dist)<Tolerance then begin
      fPlayer:=Pos;
      Pos:=fClick;
      Draw(fPlayer);
      scrRemove.X:=scrClick.X*2-scrPlayer.X;
      scrRemove.Y:=scrClick.Y*2-scrPlayer.Y;
      fRemove:=InvGetPoint(scrRemove);
      Field[fRemove.Y,fRemove.X]:=false;
      if (fRemove.X=Dest.X) AND (fRemove.Y=Dest.Y) then begin
        FindDest;
        Draw(Dest);
      end;
      Draw(fRemove);
      if (Pos.X=Dest.X) AND (Pos.Y=Dest.Y) then begin
        Timer1.Enabled:=false;
        DrawWonText;
      end;
    end;
  end;
  Draw(fClick);
end;

function TMainForm.InvGetPoint(P:TPoint):TPoint;
var X1,X2,Y1,Y2:integer;
    CX,CY:integer;
    D:TPoint;
    Dist,BestDist:integer;
begin
  CX:=(P.X-BaseX) DIV (2*Ssin60);
  CY:=(P.Y-BaseY) DIV (Scos60+Sbase);
  X1:=CX; X2:=CX+1; Result.X:=CX;
  Y1:=CY; Y2:=CY+1; Result.Y:=CY;
  if X1<0 then X1:=0; if X2>MaxX then X2:=MaxX;
  if Y1<0 then Y1:=0; if Y2>MaxY then Y2:=MaxY;

  BestDist:=MAXINT;
  for CY:=Y1 to Y2 do
    for CX:=X1 to X2 do begin
      D:=GetPoint(Point(CX,CY));
      Dist:=sqr(D.X-P.X)+sqr(D.Y-P.Y);
      if Dist<BestDist then begin
        Result.X:=CX; Result.Y:=CY;
        BestDist:=Dist;
      end;
    end;
end;

procedure TMainForm.BRestartClick(Sender: TObject);
begin
  MakeField;
  PaintBox1.Invalidate;
  Timer1.Enabled:=true;
end;

procedure TMainForm.CTrapsClick(Sender: TObject; Button: TUDBtnType);
begin
  Traps:=CTraps.Position;
end;

procedure TMainForm.Timer1Timer(Sender: TObject);
begin
  Draw(Point(1+random(MaxX-1),1+random(MaxY-1)));
end;

procedure TMainForm.DrawWonText;
var TextSize:TSize;
    I,BX,BY:integer;
begin
  with PaintBox1.Canvas do begin
    Brush.Style:=bsClear;
    Font.Name:='Verdana';
    Font.Style:=[fsBold,fsItalic];
    Font.Size:=WinTextSize;
    TextSize:=TextExtent(WinText);
    BX:=(252-TextSize.CX)SHR 1;
    BY:=(252-TextSize.CY)SHR 1;
    for I:=5 downto 1 do begin
      Font.Color:=$002800*(5-I);
      TextOut(BX+I,BY+I,WinText);
    end;
    Font.Color:=$00FF00;
    TextOut(BX,BY,WinText);
    Brush.Style:=bsSolid;
  end;
end;

end.
